"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NatsJetStreamServer = void 0;
const microservices_1 = require("@nestjs/microservices");
const nats_1 = require("nats");
const rxjs_1 = require("rxjs");
const constants_1 = require("./constants");
const nats_jetstream_context_1 = require("./nats-jetstream.context");
const server_consumer_options_builder_1 = require("./utils/server-consumer-options-builder");
class NatsJetStreamServer extends microservices_1.Server {
    constructor(options) {
        super();
        this.options = options;
        this.transportId = constants_1.NATS_JETSTREAM_TRANSPORT;
        this.codec = (0, nats_1.JSONCodec)();
    }
    async listen(callback) {
        if (!this.nc) {
            this.nc = await (0, nats_1.connect)(this.options.connectionOptions);
            if (this.options.connectionOptions.connectedHook) {
                this.options.connectionOptions.connectedHook(this.nc);
            }
        }
        this.jsm = await this.nc.jetstreamManager(this.options.jetStreamOptions);
        if (this.options.streamConfig) {
            await this.setupStream();
        }
        await this.bindEventHandlers();
        this.bindMessageHandlers();
        callback();
    }
    async close() {
        await this.nc.drain();
        this.nc = undefined;
    }
    async bindEventHandlers() {
        const eventHandlers = [...this.messageHandlers.entries()].filter(([, handler]) => handler.isEventHandler);
        const js = this.nc.jetstream(this.options.jetStreamOptions);
        for (const [subject, eventHandler] of eventHandlers) {
            const consumerOptions = (0, server_consumer_options_builder_1.serverConsumerOptionsBuilder)(this.options.consumerOptions, subject);
            const subscription = await js.subscribe(subject, consumerOptions);
            this.logger.log(`Subscribed to ${subject} events`);
            const done = (async () => {
                for await (const msg of subscription) {
                    try {
                        const data = this.codec.decode(msg.data);
                        const context = new nats_jetstream_context_1.NatsJetStreamContext([msg]);
                        const resultOrStream = await eventHandler(data, context);
                        if ((0, rxjs_1.isObservable)(resultOrStream)) {
                            const connectableSource = (0, rxjs_1.connectable)(resultOrStream, {
                                connector: () => new rxjs_1.Subject(),
                                resetOnDisconnect: false,
                            });
                            connectableSource.connect();
                        }
                    }
                    catch (err) {
                        this.logger.error(err.message, err.stack);
                        msg.term();
                    }
                }
            })();
            done.then(() => {
                if (this.nc.isDraining() || this.nc.isClosed()) {
                    return;
                }
                subscription.destroy();
                this.logger.log(`Unsubscribed ${subject}`);
            });
        }
    }
    bindMessageHandlers() {
        const messageHandlers = [...this.messageHandlers.entries()].filter(([, handler]) => !handler.isEventHandler);
        for (const [subject, messageHandler] of messageHandlers) {
            const subscriptionOptions = {
                queue: this.options.consumerOptions.deliverTo,
                callback: async (err, msg) => {
                    if (err) {
                        return this.logger.error(err.message, err.stack);
                    }
                    const payload = this.codec.decode(msg.data);
                    const context = new nats_jetstream_context_1.NatsContext([msg]);
                    const response$ = this.transformToObservable(messageHandler(payload, context));
                    this.send(response$, (response) => msg.respond(this.codec.encode(response)));
                },
            };
            this.nc.subscribe(subject, subscriptionOptions);
            this.logger.log(`Subscribed to ${subject} messages`);
        }
    }
    async setupStream() {
        const { streamConfig } = this.options;
        const streams = await this.jsm.streams.list().next();
        const reqStreamConfigs = !Array.isArray(streamConfig)
            ? [streamConfig]
            : streamConfig;
        for (const requiredStreamConfig of reqStreamConfigs) {
            const stream = streams.find((stream) => stream.config.name === requiredStreamConfig.name);
            if (stream) {
                const streamSubjects = new Set([
                    ...stream.config.subjects,
                    ...requiredStreamConfig.subjects,
                ]);
                const streamInfo = await this.jsm.streams.update(stream.config.name, {
                    ...stream.config,
                    ...requiredStreamConfig,
                    subjects: [...streamSubjects.keys()],
                });
                this.logger.log(`Stream ${streamInfo.config.name} updated`);
            }
            else {
                const streamInfo = await this.jsm.streams.add(requiredStreamConfig);
                this.logger.log(`Stream ${streamInfo.config.name} created`);
            }
        }
    }
}
exports.NatsJetStreamServer = NatsJetStreamServer;
//# sourceMappingURL=server.js.map